# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from tico.quantization.config.gptq import GPTQConfig


@dataclass
class FPIGPTQConfig(GPTQConfig):
    """
    Configuration for FPIGPTQ (Fixed Point Iteration).
    """

    def __init__(self, verbose: bool = False, show_progress: bool = True):
        self.verbose = verbose
        self.show_progress = show_progress

    @property
    def name(self) -> str:
        return "fpi_gptq"
